
## library
library(fda)
library(fields)
library(SpatialTools)
library(Iso)
library(qvalue)
library(rlang)
library(sandwich)
library(lmtest)
library(geoR)

# Generating covariates and grid 

L = 42     
N = L^2           
grid = seq(0, 1, length = L)
xy = as.matrix( expand.grid(grid, grid) )

# Main functions
# AR: Generate VAR process
## Generate AR process
AR <- function(T, dim, rho, G)
{
  data = matrix(0, dim, T+200)    
  for (i in 1:(T+199)) data[,i+1] = rho*data[,i] +  G%*%rnorm(dim)*sqrt(1-rho^2)
  return( data[, 201:(T+200)] ) 
}

## Calculate self-normalizaiton test statistic
SN <- function(zz)
{
  T = length(zz)
  cmean = cumsum(zz)/(1:T)
  stat = T*mean(zz)^2/sum((cmean - cmean[T])^2*(1:T/T)^2)
  return(stat)
}

load('/home/syun13/Multiple_NewSim/CV_SelfNorm.Rdata')  ## Call test statistics


###########################################################################################################
### Nonparametric mirror procedure function (BH and JS as references)
###########################################################################################################
NonparamL =function(p_value, delta.tr, basis.tr,init.param, Lik=TRUE) { 
  
  # FDR control using nonparametric estimate for f1
  pval = p_value
  out = sort(pval, index.return=TRUE)
  pval = out$x
  ind = out$ix
  basis = basis.tr[ind,]
  delta = delta.tr[ind]		
  df = c(pval[1],diff(pval))       # z(i)-z(i-1)
  f1.int = f1 = dexp(pval,200)     # initial yi=f1(z)
  pi0  = runif(N,0,1)		  # initial pi(s)
  f0 = rep(1,N)		        # known   f0(zi)
  
  ### ESTIMATION METHOD
  q = pi0*f0/(pi0*f0+(1-pi0)*f1);Nm = d = ncol(basis)
  beta.int = beta.cur = init.param  # initial value
  K = 100
  lik = -1e5
  iter = 1
  epsilon = 1
  while(iter <= K & epsilon > 1e-9)
  {
    # E-step: 
    q = pi0*f0/(pi0*f0+(1-pi0)*f1)
    
    # M-step (Newton method):
    Q = function(beta) {
      sum(-q*(basis %*% beta)-log(1/(exp(basis %*% beta)+1)))
    }
    grr = function(beta) {
      A = matrix(-q, N, Nm)*basis 
      C = exp(basis %*% beta)
      C[C==Inf] = 1e+30
      B = basis*matrix(C/(C+1), N, Nm)
      return(apply(A+B,2,sum))
    }
    fit.optim = nlminb(beta.int, Q, gradient=grr)   ### Initial value !!!
    beta.cur = fit.optim$par
    pi0 = plogis(basis %*% beta.cur)
    
    # PAVA
    q1 = 1-q
    qsum = sum(q1)
    y = -df*qsum/q1
    
    ### modify the code slightly
    y[is.na(y)] = -Inf
    y[y==0] = max(y[y!=0])
    y[y==-Inf] = min(y[y!=-Inf])
    
    out = pava(y, q1, decreasing=TRUE, long.out=FALSE, stepfun=FALSE)
    f1 = -1/out 
    f1 = f1/sum( f1*df )
    
    iter = iter + 1
    lik.cur = sum(log(pi0*f0+(1-pi0)*f1))
    epsilon = (lik.cur - lik[length(lik)])/abs(lik[length(lik)])
    lik = c(lik, lik.cur)
    if(iter%%200==0) print(iter)
  }
  # Local approach
  lfdr <- pi0*f0/(pi0*f0+(1-pi0)*f1)
  k <- sum(cumsum(sort(lfdr))/(1:(N)) < 0.1)               # Nominal level 10%
  ind1 <- sort(lfdr, index.return=TRUE)$ix[1:k]
  sum(delta == 0)/length(delta)
  FDP.1 <- mean(delta[ind1] == 0)             # FDP (should be close to the nominal level)
  Power.1 <- sum(delta[ind1]) / (sum(delta==1))            # Power (the larger the better) 
  
  # Shape-adaptive approach
  f.step = stepfun(pval, c(f1[1],f1), right = FALSE) 
  fp = f.step(pval)
  f1p = f.step(1-pval)
  nu = (1-pi0)/pi0*f1p
  de = (1-pi0)/pi0*fp
  TT = sort(c(nu,de))
  i = 1
  hit = FALSE
  while(i <= length(TT) & hit == FALSE)
  {
    if ( sum(nu > TT[i])/max(1,sum(de > TT[i])) < 0.1 & sum(de > TT[i])>0) hit=TRUE
    i = i+1
  }
  if(hit == TRUE) {
    FDP.2 = mean( delta[(1:N)[de > TT[i]]] ==0 )
    Power.2 = sum( delta[(1:N)[de > TT[i]]] ==1 )/(sum(delta==1))  
  } else
  {
    FDP.2 = 0;
    Power.2 = 0;
  }  
  
  # John Storey approach
  
  p.adj = qvalue(pval, fdr.level=0.1)
  indQ = which(p.adj$significant == TRUE) 	
  FDP.J = mean(delta[indQ] == 0)                   
  Power.J = sum(delta[indQ]) / (sum(delta == 1))
  
  # BH approach
  
  p1.adj = p.adjust(pval, method='BH', n =length(pval))
  indBH = which(p1.adj <0.10)
  
  FDP.B <- mean(delta[indBH] == 0)                    # FDP (should be close to the nominal level)
  Power.B <- sum(delta[indBH]) / (sum(delta==1));
  
  if (Lik == FALSE) return(list(FDP.L = FDP.1, FDP.A = FDP.2, Power.L = Power.1, Power.A = Power.2
                                , FDP.J = FDP.J, FDP.B = FDP.B, Power.B = Power.B, Power.J = Power.J))
  #L = fit.optim$objective
  L = lik[length(lik)]
  if (Lik == TRUE) return(list(E = epsilon, L = L))
}


####################################################
## SET up ## 
####################################################
# (1) Basis function
# (2) delta(s) & pi(s)
# (3) mean function
# (4) Covariance matrix

# (1) Basis function
basis.tr = cbind(1, xy, xy[,1]^2, xy[,2]^2,xy[,1]*xy[,2],xy[,1]^2*xy[,2],xy[,1]*xy[,2]^2, xy[,1]^3, xy[,2]^3,(xy[,1]*xy[,2])^2)  # Use spatial locations as covariates
beta.tr = c(-2, 3, 3, 1, -1, 2, -2, 2, -3, 2.5, 1)  # Null proportion 70%
#beta.tr = c(-1, 3, 2, 1, -1, 2, -2, 2, -3, 3, 2) # Null proportion 80%
#beta.tr = c(-0.5, 3, 3, 2, -1, 2, -2, 2, -3, 3, 3) # Null proportion 90%
Nm = dim(basis.tr)[2]

# (2) delta(s) & pi(s) 
eta = basis.tr %*% beta.tr	   
mean(exp(eta)/(1+exp(eta)))
pi.true = exp(eta)/(1+exp(eta))

set.seed(100)
delta.tr = rbinom(N, size=1, prob=1-pi.true)

# (3) mean function
effect = 0.2
mu = basis.tr %*% rep(effect, Nm)
yy.mean = as.vector(mu*delta.tr)

# (4) Covariance matrix
T=500
distmat <- as.matrix(dist(xy))  ## compuate distance matrix
sigma2 = 0.2; phi=0.1; sig1=0.3; sig2 = 0.5; rho=0.3

covm <- sigma2*exp(-distmat/phi)  ## compute covariance matrix
Dchol = t(chol(covm))

## Try different set of inital values 
Nm = 11; N = 1764
L=0.1
param = c()
for ( kk in 1:100) { 
  U = kk %/% 10 + 1
  set.seed(kk)
  param = rbind(param, runif(Nm, L, U))
}

###########################################################################################
## Parallel calculation 
###########################################################################################

set.seed(NULL)
library(doParallel)
registerDoParallel(cores=12)
detectCores()

comb <- function(x, ...) {
  lapply(seq_along(x),
         function(i) c(x[[i]], lapply(list(...), function(y) y[[i]])))
}

s = Sys.time()

ptime <- system.time({
  r <- foreach(kk=1:100,.combine='comb',.multicombine=TRUE
               ,.init=list(list(), list(), list(), list(), list(), list(), list())
               , .packages = c("fda", "fields", "SpatialTools", "Iso", "qvalue"
                               , "sandwich", "lmtest")) %dopar% {
                                 Res.FDP = Res.Power = c()
                                 
                                 ### (1) Std Normal 
                                 x.err = matrix(rnorm(N*T,0,1), N, T)
                                 y.err = matrix(rnorm(N*T,0,1), N, T)
                                 
                                 xx1 = x.err + AR(T, N, rho, diag(1,N))
                                 yy1 = y.err + AR(T, N, rho, diag(1,N)) + yy.mean	
                                 
                                 zz1 = xx1 - yy1
                                 
                                 ### (2) Gaussian
                                 
                                 D.Gaussx = D.Gaussy = c()
                                 
                                 for ( m in 1:T) { 
                                   Dtmpx <- Dchol %*%rnorm(N);Dtmpy <- Dchol %*%rnorm(N)
                                   D.Gaussx = cbind(D.Gaussx, Dtmpx);D.Gaussy = cbind(D.Gaussy, Dtmpy)
                                 }
                                 xx2 = D.Gaussx + AR(T, N, rho, diag(1,N))
                                 yy2 = yy.mean + D.Gaussy + AR(T, N, rho, diag(1,N))
                                 zz2 = xx2 - yy2
                                 
                                 ### (3) Random Effect model1
                                 
                                 x.ef = matrix(rnorm(Nm*T,0,sig1), Nm, T)
                                 y.ef = matrix(rnorm(Nm*T,0,sig1), Nm, T)
                                 
                                 xx3 = basis.tr %*% x.ef + AR(T, N, rho, diag(1,N)) 
                                 yy3 = basis.tr %*% y.ef + AR(T, N, rho, diag(1,N)) + yy.mean
                                 zz3 = xx3 - yy3
                                 
                                 ### (4) Random Effect model2
                                 
                                 x.ef = matrix(rnorm(Nm*T,0,sig2), Nm, T)
                                 y.ef = matrix(rnorm(Nm*T,0,sig2), Nm, T)
                                 
                                 xx4 = basis.tr %*% x.ef + AR(T, N, rho, diag(1,N)) 
                                 yy4 = basis.tr %*% y.ef + AR(T, N, rho, diag(1,N)) + yy.mean
                                 zz4 = xx4 - yy4
                                 
                                 
                                 ## P-value
                                 pval1 = pval2 = pval3 = pval4 = rep(NA, N)
                                 for(j in 1:N)  {
                                   stat1 = SN(zz1[j,]);stat2 = SN(zz2[j,]);stat3 = SN(zz3[j,]); stat4 = SN(zz4[j,])
                                   pval1[j] = 1-sn.cdf(stat1); pval2[j] = 1-sn.cdf(stat2) ; pval3[j] = 1-sn.cdf(stat3); pval4[j] = 1-sn.cdf(stat4)
                                 }
                                 
                                 ## Choose the best initial values ## 
                                 Res.L1 = Res.L2 = Res.L3 = Res.L4 = c()
                                 for ( j in 1:100) { 
                                   tmp1  =NonparamL(pval1, delta.tr, basis.tr, param[j,],Lik=TRUE)
                                   tmp2 = NonparamL(pval2, delta.tr, basis.tr, param[j,],Lik=TRUE)
                                   tmp3 = NonparamL(pval3, delta.tr, basis.tr, param[j,],Lik=TRUE)
                                   tmp4 = NonparamL(pval4, delta.tr, basis.tr, param[j,],Lik=TRUE)
                                   Res.L1 = c(Res.L1,tmp1$L); Res.L2 = c(Res.L2, tmp2$L); Res.L3 = c(Res.L3, tmp3$L)
                                   ; Res.L4 = c(Res.L4, tmp4$L)
                                   if(j%%10==0) print(j)
                                 }
                                 id1 = which.max(Res.L1); id2 = which.max(Res.L2); id3 = which.max(Res.L3); id4 = which.max(Res.L4)
                                 max.id = c(id1, id2, id3, id4)
                                 
                                 R1 = NonparamL(pval1, delta.tr, basis.tr, param[id1,],Lik=FALSE)
                                 R2 = NonparamL(pval2, delta.tr, basis.tr, param[id2,],Lik=FALSE)
                                 R3 = NonparamL(pval3, delta.tr, basis.tr, param[id3,],Lik=FALSE)
                                 R4 = NonparamL(pval4, delta.tr, basis.tr, param[id4,],Lik=FALSE)
                                 
                                 res.FDP = c(R1$FDP.L, R1$FDP.A, R1$FDP.B, R1$FDP.J, R2$FDP.L, R2$FDP.A, R2$FDP.B, R2$FDP.J
                                             , R3$FDP.L, R3$FDP.A, R3$FDP.B, R3$FDP.J, R4$FDP.L, R4$FDP.A, R4$FDP.B, R4$FDP.J)
                                 
                                 res.Power = c(R1$Power.L, R1$Power.A, R1$Power.B, R1$Power.J, R2$Power.L, R2$Power.A, R2$Power.B, R2$Power.J
                                               , R3$Power.L, R3$Power.A, R3$Power.B, R3$Power.J, R4$Power.L, R4$Power.A, R4$Power.B, R4$Power.J)
                                 
                                 list(pval1, pval2, pval3,pval4, max.id, res.FDP, res.Power)
                                 
                               }
})
e = Sys.time(); e-s


save(delta.tr, basis.tr, r, file='NonparamMirror_result.Rdata')

